/* $Id: api.c 119 2023-06-17 07:16:11Z nishi $ */
#include <swinehub/api.h>

#include <swinehub.h>
#include <swinehub/config.h>
#include <swinehub/util.h>
#include <swinehub/sha512.h>
#include <swinehub/language.h>
#include <swinehub/db.h>

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <dirent.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <unistd.h>
#include <fcntl.h>
#include <limits.h>

#include <openssl/sha.h>

#define WRITE(fd, data) write(fd, data, strlen(data))
#define _FWRITE(f, data) fwrite(data, strlen(data), 1, f)

EXTERN_DB;

extern char* swinehub_body;

void swinehub_api(struct swinehub_config* config){
	char* path = getenv("PATH_REPLACED");
	if(strcmp(path, "/v1/repo/latest-log") == 0){
		printf("Content-Type: text/plain\r\n");
	}else{
		printf("Content-Type: application/json\r\n");
	}
	if(strcmp(path, "/v1") == 0){
		printf("Status: 200 OK\r\n");
		printf("\r\n");
		printf("{\"message\": \"API is working\", \"code\": %d}", SWINEHUB_CODE_OK);
	}else if(strcmp(path, "/v1/version") == 0){
		printf("Status: 200 OK\r\n");
		printf("\r\n");
		printf("{\"version\": \"%s\", \"code\": %d}", SWINEHUB_VERSION, SWINEHUB_CODE_OK);
	}else if(strcmp(path, "/v1/valid") == 0){
		datum key;
		key.dptr = swinehub_body;
		key.dsize = strlen(swinehub_body) + 1;
		datum result = dbm_fetch(swinehub_token_db, key);
		if(result.dptr != NULL){
			printf("Status: 200 OK\r\n");
			printf("\r\n");
			printf("{\"message\": \"Valid token\", \"username\": \"%s\", \"code\": %d}", (char*)result.dptr, SWINEHUB_CODE_OK);
		}else{
			printf("Status: 403 Forbidden\r\n");
			printf("\r\n");
			printf("{\"error\": \"Invalid token\", \"code\": %d}", SWINEHUB_CODE_INVALID_TOKEN);
		}
	}else if(strcmp(path, "/v1/repos") == 0){
		datum key;
		key.dptr = swinehub_body;
		key.dsize = strlen(swinehub_body) + 1;
		datum result = dbm_fetch(swinehub_user_db, key);
		if(result.dptr != NULL){
			printf("Status: 200 OK\r\n");
			printf("\r\n");
			char* repos = malloc(1);
			repos[0] = 0;
			DIR* dir = opendir(config->path);
			struct dirent* d;
			char* prefix = sstrcat(swinehub_body, "-");
			while((d = readdir(dir)) != NULL){
				int i;
				char cond = 1;
				for(i = 0; d->d_name[i] != 0 && prefix[i] != 0; i++){
					if(prefix[i] != d->d_name[i]){
						cond = 0;
						break;
					}
				}
				if(cond){
					char* tmp = repos;
					repos = sstrcat5(tmp, repos[0] == 0 ? "" : ", ", "\"", d->d_name + strlen(prefix), "\"");
					free(tmp);
				}
			}
			free(prefix);
			printf("{\"repos\": [%s], \"code\": %d}", repos, SWINEHUB_CODE_OK);
		}else{
			printf("Status: 404 Not Found\r\n");
			printf("\r\n");
			printf("{\"error\": \"User does not exist\", \"code\": %d}", SWINEHUB_CODE_USER_DOES_NOT_EXIST);
		}
	}else if(strcmp(path, "/v1/logout") == 0){
		datum key;
		key.dptr = swinehub_body;
		key.dsize = strlen(swinehub_body) + 1;
		int err = dbm_delete(swinehub_token_db, key);
		swinehub_db_init();
		if(err == 0){
			printf("Status: 200 OK\r\n");
			printf("\r\n");
			printf("{\"message\": \"Logout successful\", \"code\": %d}", SWINEHUB_CODE_OK);
		}else{
			printf("Status: 403 Forbidden\r\n");
			printf("\r\n");
			printf("{\"error\": \"Invalid token\", \"code\": %d}", SWINEHUB_CODE_INVALID_TOKEN);
		}
		fflush(stdout);
	}else if(strcmp(path, "/v1/signup") == 0){
		if(!config->signup){
			printf("Status: 403 Forbidden\r\n");
			printf("\r\n");
			printf("{\"error\": \"Signup disabled\", \"code\": %d}", SWINEHUB_CODE_SIGNUP_DISABLED);
			return;
		}
		int i;
		char* backup = strdup(swinehub_body);
		for(i = 0; swinehub_body[i] != 0; i++){
			if(swinehub_body[i] == ':'){
				swinehub_body[i] = 0;
				break;
			}
		}
		char* username = swinehub_body;
		char* password = swinehub_body + i + 1;
		if(strlen(username) == strlen(backup)){
			free(backup);
			printf("Status: 400 Bad Request\r\n");
			printf("\r\n");
			printf("{\"error\": \"Required parameters missing, `username:password`\", \"code\": %d}", SWINEHUB_CODE_PARAM_MISSING);
			return;
		}
		if(strlen(username) > SWINEHUB_USERNAME_MAX){
			free(backup);
			printf("Status: 403 Forbidden\r\n");
			printf("\r\n");
			printf("{\"error\": \"Username too long (>%d)\", \"code\": %d}", SWINEHUB_USERNAME_MAX, SWINEHUB_CODE_USERNAME_TOO_LONG);
			return;
		}
		int hi;
		for(hi = 0; username[hi] != 0; hi++){
			if(username[hi] == '-'){
				printf("Status: 400 Bad Request\r\n");
				printf("\r\n");
				printf("{\"error\": \"Username does not accept hyphens\", \"code\": %d}", SWINEHUB_CODE_NO_HYPHENS);
				return;
			}else if(username[hi] == '#'){
				printf("Status: 400 Bad Request\r\n");
				printf("\r\n");
				printf("{\"error\": \"Username does not accept hashes\", \"code\": %d}", SWINEHUB_CODE_NO_HASHES);
				return;
			}else if(username[hi] == '['){
				printf("Status: 400 Bad Request\r\n");
				printf("\r\n");
				printf("{\"error\": \"Username does not accept brackets\", \"code\": %d}", SWINEHUB_CODE_NO_BRACKETS);
				return;
			}else if(username[hi] == ']'){
				printf("Status: 400 Bad Request\r\n");
				printf("\r\n");
				printf("{\"error\": \"Username does not accept brackets\", \"code\": %d}", SWINEHUB_CODE_NO_BRACKETS);
				return;
			}else if(username[hi] == '='){
				printf("Status: 400 Bad Request\r\n");
				printf("\r\n");
				printf("{\"error\": \"Username does not accept equals\", \"code\": %d}", SWINEHUB_CODE_NO_EQUALS);
				return;
			}
		}
		free(backup);
		datum key;
		key.dptr = username;
		key.dsize = strlen(username) + 1;
		datum result = dbm_fetch(swinehub_user_db, key);
		if(result.dptr != NULL){
			printf("Status: 409 Conflict\r\n");
			printf("\r\n");
			struct swinehub_user* usr = (struct swinehub_user*)result.dptr;
			printf("{\"error\": \"User exists\", \"code\": %d}", SWINEHUB_CODE_USER_EXISTS);
		}else{
			key.dptr = username;
			key.dsize = strlen(username) + 1;
			datum value;
			struct swinehub_user user;
			memset(user.username, 0, SWINEHUB_USERNAME_MAX + 1);
			strcpy(user.username, username);
			user.isteam = 0;
			unsigned char* sha512passwd = swinehub_sha512(password);
			memcpy(user.password, sha512passwd, SHA512_DIGEST_LENGTH);
			free(sha512passwd);
			value.dptr = (void*)&user;
			value.dsize = sizeof(user);
			if(0 != dbm_store(swinehub_user_db, key, value, DBM_REPLACE)){
				printf("Status: 500 Internal Server Error\r\n");
				printf("\r\n");
				printf("{\"message\": \"Database update failed: %s\", \"code\": %d}", strerror(errno), SWINEHUB_CODE_DB);
			}else{
				int pid = fork();
				if(pid == 0){
					execl(config->htpasswd, config->htpasswd, "-b", SWINEHUB_AUTH, user.username, password, NULL);
					_exit(-1);
				}else{
					waitpid(pid, 0, 0);
				}
				printf("Status: 200 OK\r\n");
				printf("\r\n");
				printf("{\"message\": \"Database successfully updated\", \"code\": %d}", SWINEHUB_CODE_OK);
			}
			swinehub_db_init();
		}
		fflush(stdout);
	}else if(strcmp(path, "/v1/team/create") == 0){
		if(!config->signup){
			printf("Status: 403 Forbidden\r\n");
			printf("\r\n");
			printf("{\"error\": \"Signup disabled\", \"code\": %d}", SWINEHUB_CODE_SIGNUP_DISABLED);
			return;
		}
		int i;
		char* backup = strdup(swinehub_body);
		int index = -1;
		int index2 = -1;
		for(i = 0; swinehub_body[i] != 0; i++){
			if(swinehub_body[i] == ':'){
				swinehub_body[i] = 0;
				if(index == -1){
					index = i;
				}else if(index2 == -1){
					index2 = i;
				}
			}
		}
		char* username = swinehub_body;
		char* password = swinehub_body + index + 1;
		char* users = index2 == -1 ? NULL : swinehub_body + index2 + 1;
		if(strlen(username) == strlen(backup)){
			free(backup);
			printf("Status: 400 Bad Request\r\n");
			printf("\r\n");
			printf("{\"error\": \"Required parameters missing, `teamname:password:users`\", \"code\": %d}", SWINEHUB_CODE_PARAM_MISSING);
			return;
		}
		if(index2 == -1){
			free(backup);
			printf("Status: 400 Bad Request\r\n");
			printf("\r\n");
			printf("{\"error\": \"Required parameters missing, `teamname:password:users`\", \"code\": %d}", SWINEHUB_CODE_PARAM_MISSING);
			return;
		}
		if(strlen(username) > SWINEHUB_USERNAME_MAX){
			free(backup);
			printf("Status: 403 Forbidden\r\n");
			printf("\r\n");
			printf("{\"error\": \"Teamname too long (>%d)\", \"code\": %d}", SWINEHUB_USERNAME_MAX, SWINEHUB_CODE_USERNAME_TOO_LONG);
			return;
		}
		int hi;
		char* _users = users;
		users = malloc(1);
		users[0] = 0;
		char del = 0;
		char* busers = strdup(_users);
		int size_users = 0;
		for(i = strlen(busers) - 1; i != -1; i--){
			if(busers[i] == ' ' && del == 1){
				busers[i] = 1;
				continue;
			}
			del = 0;
			if(busers[i] == ','){
				del = 1;
			}
		}
		del = 0;
		for(i = 0; busers[i] != 0; i++){
			if(busers[i] == ' ' && del == 1){
				busers[i] = 1;
				continue;
			}
			del = 0;
			if(busers[i] != 1){
				users = realloc(users, size_users + 2);
				users[size_users] = busers[i];
				users[size_users + 1] = 0;
				size_users++;
			}
			if(busers[i] == ','){
				del = 1;
			}
		}
		free(busers);
		for(hi = 0; username[hi] != 0; hi++){
			if(username[hi] == '-'){
				printf("Status: 400 Bad Request\r\n");
				printf("\r\n");
				printf("{\"error\": \"Teamname does not accept hyphens\", \"code\": %d}", SWINEHUB_CODE_NO_HYPHENS);
				return;
			}else if(username[hi] == '#'){
				printf("Status: 400 Bad Request\r\n");
				printf("\r\n");
				printf("{\"error\": \"Teamname does not accept hashes\", \"code\": %d}", SWINEHUB_CODE_NO_HASHES);
				return;
			}else if(username[hi] == '['){
				printf("Status: 400 Bad Request\r\n");
				printf("\r\n");
				printf("{\"error\": \"Teamname does not accept brackets\", \"code\": %d}", SWINEHUB_CODE_NO_BRACKETS);
				return;
			}else if(username[hi] == ']'){
				printf("Status: 400 Bad Request\r\n");
				printf("\r\n");
				printf("{\"error\": \"Teamname does not accept brackets\", \"code\": %d}", SWINEHUB_CODE_NO_BRACKETS);
				return;
			}else if(username[hi] == '='){
				printf("Status: 400 Bad Request\r\n");
				printf("\r\n");
				printf("{\"error\": \"Teamname does not accept equals\", \"code\": %d}", SWINEHUB_CODE_NO_EQUALS);
				return;
			}
		}
		free(backup);
		datum key;
		key.dptr = username;
		key.dsize = strlen(username) + 1;
		datum result = dbm_fetch(swinehub_user_db, key);
		if(result.dptr != NULL){
			printf("Status: 409 Conflict\r\n");
			printf("\r\n");
			struct swinehub_user* usr = (struct swinehub_user*)result.dptr;
			printf("{\"error\": \"Team exists\", \"code\": %d}", SWINEHUB_CODE_USER_EXISTS);
		}else{
			key.dptr = username;
			key.dsize = strlen(username) + 1;
			datum value;
			struct swinehub_user user;
			memset(user.username, 0, SWINEHUB_USERNAME_MAX + 1);
			strcpy(user.username, username);
			user.isteam = 1;
			unsigned char* sha512passwd = swinehub_sha512(password);
			memcpy(user.password, sha512passwd, SHA512_DIGEST_LENGTH);
			free(sha512passwd);
			value.dptr = (void*)&user;
			value.dsize = sizeof(user);
			if(0 != dbm_store(swinehub_user_db, key, value, DBM_REPLACE)){
				printf("Status: 500 Internal Server Error\r\n");
				printf("\r\n");
				printf("{\"message\": \"Database update failed: %s\", \"code\": %d}", strerror(errno), SWINEHUB_CODE_DB);
			}else{
				int pid = fork();
				if(pid == 0){
					execl(config->htpasswd, config->htpasswd, "-b", SWINEHUB_AUTH, user.username, password, NULL);
					_exit(-1);
				}else{
					waitpid(pid, 0, 0);
				}
				char* data = malloc(1);
				data[0] = 0;
				FILE* f = fopen(SWINEHUB_AUTHZ, "r");
				flockfile(f);
				char* buf = malloc(2);
				buf[1] = 0;
				char* linebuf = malloc(1);
				linebuf[0] = 0;
				while(1){
					fread(buf, 1, 1, f);
					if(feof(f) || buf[0] == '\n'){
						if(!feof(f)){
							char* tmp = data;
							data = sstrcat3(data, linebuf, "\n");
							free(tmp);
							if(strcmp(linebuf, "#%NEW-TEAM") == 0){
								tmp = data;
								data = sstrcat4(data, "#%START-TEAM ", username, "\n");
								free(tmp);
								tmp = data;
								data = sstrcat5(data, username, " = ", users, "\n");
								free(tmp);
								tmp = data;
								data = sstrcat(data, "#%END-TEAM\n");
								free(tmp);
							}
						}
						free(linebuf);
						linebuf = malloc(1);
						linebuf[0] = 0;
						if(feof(f)) break;
					}else{
						char* tmp = linebuf;
						linebuf = sstrcat(linebuf, buf);
						free(tmp);
					}
				}
				free(linebuf);
				free(buf);
				fclose(f);
				f = fopen(SWINEHUB_AUTHZ, "w");
				_FWRITE(f, data);
				funlockfile(f);
				fclose(f);
				free(data);
				printf("Status: 200 OK\r\n");
				printf("\r\n");
				printf("{\"message\": \"Database successfully updated\", \"code\": %d}", SWINEHUB_CODE_OK);
			}
			swinehub_db_init();
		}
		fflush(stdout);
		free(users);
	}else if(strcmp(path, "/v1/team/change-users") == 0){
		char* users = malloc(1);
		users[0] = 0;
		char* busers = strdup(swinehub_body);
		int i;
		char comma = 0;
		for(i = strlen(busers) - 1; i != -1; i--){
			if(busers[i] == ' ' && comma == 1){
				busers[i] = 1;
				continue;
			}
			comma = 0;
			if(busers[i] == ','){
				comma = 1;
			}
		}
		comma = 0;
		int index = 0;
		for(i = 0; busers[i] != 0; i++){
			if(busers[i] == ' ' && comma == 1){
				continue;
			}
			comma = 0;
			if(busers[i] == ','){
				comma = 1;
			}
			if(busers[i] != 1){
				users = realloc(users, index + 2);
				users[index] = busers[i];
				users[index + 1] = 0;
				index++;
			}
		}
		free(busers);
		datum key;
		key.dptr = getenv("HTTP_X_TOKEN");
		key.dsize = strlen(getenv("HTTP_X_TOKEN")) + 1;
		datum result = dbm_fetch(swinehub_token_db, key);
		if(result.dptr != NULL){
			FILE* f = fopen(SWINEHUB_AUTHZ, "r");
			flockfile(f);
			char* readbuf = malloc(2);
			readbuf[1] = 0;
			char* buf = malloc(1);
			buf[0] = 0;
			char* fullbuf = malloc(1);
			char skip = 0;
			char* find = sstrcat("#%START-TEAM ", result.dptr);
			while(1){
				fread(readbuf, 1, 1, f);
				if(readbuf[0] == '\n' || feof(f)){
					if(skip == 0){
						char* tmp = fullbuf;
						fullbuf = sstrcat3(tmp, buf, "\n");
						free(tmp);
						if(strcmp(find, buf) == 0) skip = 1;
					}else{
						char* tmp = fullbuf;
						fullbuf = sstrcat5(tmp, result.dptr, " = ", users, "\n");
						free(tmp);
						skip = 0;
					}
					free(buf);
					buf = malloc(1);
					buf[0] = 0;
					if(feof(f)) break;
				}else{
					char* tmp = buf;
					buf = sstrcat(buf, readbuf);
					free(tmp);
				}
			}
			free(find);
			free(buf);
			free(readbuf);
			funlockfile(f);
			fclose(f);
			f = fopen(SWINEHUB_AUTHZ, "w");
			flockfile(f);
			fwrite(fullbuf, strlen(fullbuf), 1, f);
			funlockfile(f);
			fclose(f);
			free(fullbuf);
			printf("Status: 200 OK\r\n");
			printf("\r\n");
			printf("{\"message\": \"Updated successfully\", \"code\": %d}", SWINEHUB_CODE_OK);
		}else{
			printf("Status: 403 Forbidden\r\n");
			printf("\r\n");
			printf("{\"error\": \"Invalid token\", \"code\": %d}", SWINEHUB_CODE_INVALID_TOKEN);
		}
		free(users);
	}else if(strcmp(path, "/v1/login") == 0){
		int i;
		char* backup = strdup(swinehub_body);
		for(i = 0; swinehub_body[i] != 0; i++){
			if(swinehub_body[i] == ':'){
				swinehub_body[i] = 0;
				break;
			}
		}
		char* username = swinehub_body;
		char* password = swinehub_body + i + 1;
		if(strlen(username) == strlen(backup)){
			free(backup);
			printf("Status: 400 Bad Request\r\n");
			printf("\r\n");
			printf("{\"error\": \"Required parameters missing, `username:password`\", \"code\": %d}", SWINEHUB_CODE_PARAM_MISSING);
			return;
		}
		datum key;
		key.dptr = username;
		key.dsize = strlen(username) + 1;
		datum result = dbm_fetch(swinehub_user_db, key);
		if(result.dptr == NULL){
			printf("Status: 404 Not Found\r\n");
			printf("\r\n");
			printf("{\"error\": \"User does not exist\", \"code\": %d}", SWINEHUB_CODE_USER_DOES_NOT_EXIST);
		}else{
			struct swinehub_user* usr = (struct swinehub_user*)result.dptr;
			char* passwd = (char*)swinehub_sha512(password);
			if(memcmp(usr->password, passwd, SHA512_DIGEST_LENGTH) == 0){
				char* token;
				char* buf;
				FILE* f;
				int i;
				while(1){
					token = malloc(1);
					token[0] = 0;
					buf = malloc(3);
					buf[2] = 0;
					f = fopen("/dev/urandom", "r");
					for(i = 0; i < 64; i++){
						unsigned char value;
						fread(&value, 1, 1, f);
						sprintf(buf, "%.2x", value);
						char* tmp = token;
						token = sstrcat(tmp, buf);
						free(tmp);
					}
					free(buf);
					fclose(f);
					key.dptr = token;
					key.dsize = strlen(token) + 1;
					result = dbm_fetch(swinehub_token_db, key);
					if(result.dptr != NULL){
						free(token);
					}else{
						break;
					}
				}
				key.dptr = token;
				key.dsize = strlen(token) + 1;
				datum value;
				value.dptr = username;
				value.dsize = strlen(username) + 1;
				if(0 != dbm_store(swinehub_token_db, key, value, DBM_REPLACE)){
					printf("Status: 500 Internal Server Error\r\n");
					printf("\r\n");
					printf("{\"message\": \"Database update failed: %s\", \"code\": %d}", strerror(errno), SWINEHUB_CODE_DB);
				}else{
					printf("Status: 200 OK\r\n");
					printf("\r\n");
					printf("{\"token\": \"%s\", \"code\": %d}", token, SWINEHUB_CODE_OK);
				}
				swinehub_db_init();
				free(token);
			}else{
				printf("Status: 403 Forbidden\r\n");
				printf("\r\n");
				printf("{\"error\": \"Incorrect Password\", \"code\": %d}", SWINEHUB_CODE_INCORRECT_PASSWORD);
			}
			free(passwd);
			fflush(stdout);
		}
	}else if(strcmp(path, "/v1/repo") == 0){
		int i;
		char* backup = strdup(swinehub_body);
		for(i = 0; swinehub_body[i] != 0; i++){
			if(swinehub_body[i] == ':'){
				swinehub_body[i] = 0;
				break;
			}
		}
		char* username = swinehub_body;
		char* reponame = swinehub_body + i + 1;
		if(strlen(username) == strlen(backup)){
			free(backup);
			printf("Status: 400 Bad Request\r\n");
			printf("\r\n");
			printf("{\"error\": \"Required parameters missing, `username:reponame`\", \"code\": %d}", SWINEHUB_CODE_PARAM_MISSING);
			return;
		}
		free(backup);
		char* tmp = sstrcat3(config->path, "/", username);
		char* str = sstrcat3(tmp, "-", reponame);
		free(tmp);
		struct stat s;
		if(stat(str, &s) == 0){
			printf("Status: 200 OK\r\n");
			printf("\r\n");
			printf("{\"message\": \"Found the repository\", \"code\": %d}", SWINEHUB_CODE_OK);
		}else{
			printf("Status: 404 Not Found\r\n");
			printf("\r\n");
			printf("{\"error\": \"Repository not found\", \"code\": %d}", SWINEHUB_CODE_REPO_NOT_FOUND);
		}
		free(str);
	}else if(strcmp(path, "/v1/repo/dir") == 0){
		int i;
		int index = -1;
		int index2 = -1;
		for(i = 0; swinehub_body[i] != 0; i++){
			if(swinehub_body[i] == ':'){
				swinehub_body[i] = 0;
				if(index == -1){
					index = i;
				}else if(index2 == -1){
					index2 = i;
					break;
				}
			}
		}
		if(index == -1 || index2 == -1){
			printf("Status: 400 Bad Request\r\n");
			printf("\r\n");
			printf("{\"error\": \"Required parameters missing, `username:reponame:path`\", \"code\": %d}", SWINEHUB_CODE_PARAM_MISSING);
			return;
		}else{
			char* username = swinehub_body;
			char* reponame = swinehub_body + index + 1;
			char* path = swinehub_body + index2 + 1;
			char* concat = sstrcat3(username, "-", reponame);
			char* repopath = sstrcat3(config->path, "/", concat);
			char* cachepath = sstrcat3(config->cachepath, "/", concat);
			struct stat s;
			if(stat(repopath, &s) == 0){
				if(stat(cachepath, &s) != 0){
					char* filepath = sstrcat("file:///", repopath);
					int pid = fork();
					if(pid == 0){
						int nullfd = open("/dev/null", O_RDWR);
						dup2(nullfd, fileno(stdout));
						dup2(nullfd, fileno(stderr));
						chdir(config->cachepath);
						execl(config->svn, config->svn, "co", filepath, NULL);
						_exit(-1);
					}else{
						waitpid(pid, 0, 0);
					}
					free(filepath);
				}
				int pid = fork();
				if(pid == 0){
					int nullfd = open("/dev/null", O_RDWR);
					dup2(nullfd, fileno(stdout));
					dup2(nullfd, fileno(stderr));
					chdir(cachepath);
					execl(config->svn, config->svn, "up", NULL);
					_exit(-1);
				}else{
					waitpid(pid, 0, 0);
				}
			}else{
				printf("Status: 404 Not Found\r\n");
				printf("\r\n");
				printf("{\"error\": \"Repository not found\", \"code\": %d}", SWINEHUB_CODE_REPO_NOT_FOUND);
				goto free_path;
			}
			char* _fullpath = sstrcat3(cachepath, "/", path);
			char* fullpath = malloc(strlen(_fullpath) + 1);
			memset(fullpath, 0, strlen(_fullpath) + 1);
			realpath(_fullpath, fullpath);
			free(_fullpath);
			char* slash = sstrcat(cachepath, "/");
			if(!(strcmp(fullpath, cachepath) == 0 || starts_with(fullpath, slash))){
				printf("Status: 400 Bad Request\r\n");
				printf("\r\n");
				printf("{\"error\": \"Path traversal detected\", \"code\": %d}", SWINEHUB_CODE_PATH_TRAVERSAL_DETECTED);
			}else{
				if(stat(fullpath, &s) == 0){
					printf("Status: 200 OK\r\n");
					printf("\r\n");
					printf("{\"code\": %d, \"files\": [\n", SWINEHUB_CODE_OK);
					fflush(stdout);
					int* pipes = malloc(sizeof(*pipes) * 2);
					pipe(pipes);
					int pid = fork();
					if(pid == 0){
						close(pipes[0]);
						int nullfd = open("/dev/null", O_RDWR);
						dup2(nullfd, fileno(stderr));
						dup2(pipes[1], fileno(stdout));
						chdir(cachepath);
						execl(config->svn, config->svn, "ls", path, NULL);
						_exit(-1);
					}else{
						close(pipes[1]);
						char ch;
						char* buf = malloc(1);
						buf[0] = 0;
						int counter = 0;
						while(1){
							int status = read(pipes[0], &ch, 1);
							if(status == 0 && kill(pid, 0) == 0) break;
							if(ch == '\n'){
								if(counter != 0){
									printf(",");
								}
								counter = 1;
								char* fullpath = sstrcat5(cachepath, "/", path, "/", buf);
								stat(fullpath, &s);
								free(fullpath);
								printf("{\"name\": \"%s\", \"size\": %ld, \"directory\": %d, \"mode\": %d}\n", buf, s.st_size, S_ISDIR(s.st_mode) ? 1 : 0, s.st_mode);
								free(buf);
								buf = malloc(1);
								buf[0] = 0;
							}else if(ch == '"'){
								char* tmp = buf;
								buf = sstrcat(buf, "\"");
								free(tmp);
							}else{
								char* buffer = malloc(2);
								buffer[0] = ch;
								buffer[1] = 0;
								char* tmp = buf;
								buf = sstrcat(buf, buffer);
								free(tmp);
								free(buffer);
							}
						}
						free(buf);
						waitpid(pid, 0, 0);
					}
					printf("]}");
					fflush(stdout);
				}else{
					printf("Status: 404 Not Found\r\n");
					printf("\r\n");
					printf("{\"error\": \"Directory not found\", \"code\": %d}", SWINEHUB_CODE_DIRECTORY_NOT_FOUND);
				}
			}
			free(slash);
free_path:
			free(cachepath);
			free(repopath);
			free(concat);
		}
	}else if(strcmp(path, "/v1/repo/cat") == 0){
		int i;
		int index = -1;
		int index2 = -1;
		for(i = 0; swinehub_body[i] != 0; i++){
			if(swinehub_body[i] == ':'){
				swinehub_body[i] = 0;
				if(index == -1){
					index = i;
				}else if(index2 == -1){
					index2 = i;
					break;
				}
			}
		}
		if(index == -1 || index2 == -1){
			printf("Status: 400 Bad Request\r\n");
			printf("\r\n");
			printf("Required parameters missing, `username:reponame:path`");
			return;
		}else{
			char* username = swinehub_body;
			char* reponame = swinehub_body + index + 1;
			char* path = swinehub_body + index2 + 1;
			char* concat = sstrcat3(username, "-", reponame);
			char* repopath = sstrcat3(config->path, "/", concat);
			char* cachepath = sstrcat3(config->cachepath, "/", concat);
			struct stat s;
			if(stat(repopath, &s) == 0){
				if(stat(cachepath, &s) != 0){
					char* filepath = sstrcat("file:///", repopath);
					int pid = fork();
					if(pid == 0){
						int nullfd = open("/dev/null", O_RDWR);
						dup2(nullfd, fileno(stdout));
						dup2(nullfd, fileno(stderr));
						chdir(config->cachepath);
						execl(config->svn, config->svn, "co", filepath, NULL);
						_exit(-1);
					}else{
						waitpid(pid, 0, 0);
					}
					free(filepath);
				}
				int pid = fork();
				if(pid == 0){
					int nullfd = open("/dev/null", O_RDWR);
					dup2(nullfd, fileno(stdout));
					dup2(nullfd, fileno(stderr));
					chdir(cachepath);
					execl(config->svn, config->svn, "up", NULL);
					_exit(-1);
				}else{
					waitpid(pid, 0, 0);
				}
			}else{
				printf("Status: 404 Not Found\r\n");
				printf("\r\n");
				printf("Repository not found");
				goto free_it;
			}
			char* _fullpath = sstrcat3(cachepath, "/", path);
			char* fullpath = malloc(strlen(_fullpath) + 1);
			memset(fullpath, 0, strlen(_fullpath) + 1);
			realpath(_fullpath, fullpath);
			free(_fullpath);
			char* slash = sstrcat(cachepath, "/");
			if(!(strcmp(fullpath, cachepath) == 0 || starts_with(fullpath, slash))){
				printf("Status: 400 Bad Request\r\n");
				printf("\r\n");
				printf("Path traversal detected");
			}else{
				if(stat(fullpath, &s) == 0){
					printf("Status: 200 OK\r\n");
					printf("\r\n");
					fflush(stdout);
					int pid = fork();
					if(pid == 0){
						int nullfd = open("/dev/null", O_RDWR);
						dup2(nullfd, fileno(stderr));
						chdir(cachepath);
						execl(config->svn, config->svn, "cat", path, NULL);
						_exit(-1);
					}else{
						waitpid(pid, 0, 0);
					}
					fflush(stdout);
				}else{
					printf("Status: 404 Not Found\r\n");
					printf("\r\n");
					printf("Directory not found");
				}
			}
			free(slash);
free_it:
			free(cachepath);
			free(repopath);
			free(concat);
		}
	}else if(strcmp(path, "/v1/repo/readme") == 0){
		int i;
		int index = -1;
		for(i = 0; swinehub_body[i] != 0; i++){
			if(swinehub_body[i] == ':'){
				swinehub_body[i] = 0;
				if(index == -1){
					index = i;
					break;
				}
			}
		}
		if(index == -1){
			printf("Status: 400 Bad Request\r\n");
			printf("\r\n");
			printf("{\"error\": \"Required parameters missing, `username:reponame`\", \"code\": %d}", SWINEHUB_CODE_PARAM_MISSING);
			return;
		}else{
			char* username = swinehub_body;
			char* reponame = swinehub_body + index + 1;
			char* concat = sstrcat3(username, "-", reponame);
			char* repopath = sstrcat3(config->path, "/", concat);
			char* cachepath = sstrcat3(config->cachepath, "/", concat);
			struct stat s;
			if(stat(repopath, &s) == 0){
				char* filepath = sstrcat("file:///", repopath);
				int pid = fork();
				if(pid == 0){
					int nullfd = open("/dev/null", O_RDWR);
					dup2(nullfd, fileno(stdout));
					dup2(nullfd, fileno(stderr));
					chdir(config->cachepath);
					execl(config->svn, config->svn, "co", filepath, NULL);
					_exit(-1);
				}else{
					waitpid(pid, 0, 0);
				}
				free(filepath);
				filepath = sstrcat(cachepath, "/README." SW_SUFFIX);
				FILE* f = fopen(filepath, "r");
				if(f == NULL){
					free(filepath);
					filepath = sstrcat(cachepath, "/README");
					f = fopen(filepath, "r");
				}
				if(f != NULL){
					printf("Status: 200 OK\r\n");
					printf("\r\n");
					fflush(stdout);
					char ch;
					while(!feof(f)){
						fread(&ch, 1, 1, f);
						printf("%c", ch);
					}
					fflush(stdout);
				}else{
					printf("Status: 404 Not Found\r\n");
					printf("\r\n");
					printf("README not found");
				}
				free(filepath);
			}else{
				printf("Status: 404 Not Found\r\n");
				printf("\r\n");
				printf("Repository not found");
			}
			free(concat);
			free(repopath);
			free(cachepath);
		}
	}else if(strcmp(path, "/v1/repo/revision") == 0){
		int i;
		char* backup = strdup(swinehub_body);
		for(i = 0; swinehub_body[i] != 0; i++){
			if(swinehub_body[i] == ':'){
				swinehub_body[i] = 0;
				break;
			}
		}
		char* username = swinehub_body;
		char* reponame = swinehub_body + i + 1;
		if(strlen(username) == strlen(backup)){
			free(backup);
			printf("Status: 400 Bad Request\r\n");
			printf("\r\n");
			printf("{\"error\": \"Required parameters missing, `username:reponame`\", \"code\": %d}", SWINEHUB_CODE_PARAM_MISSING);
			return;
		}
		free(backup);
		char* tmp = sstrcat3(config->path, "/", username);
		char* str = sstrcat3(tmp, "-", reponame);
		free(tmp);
		char* rev = sstrcat(str, "/db/current");
		FILE* f = fopen(rev, "r");
		if(f != NULL){
			struct stat s;
			stat(rev, &s);
			char* revision = malloc(s.st_size);
			revision[s.st_size - 1] = 0;
			fread(revision, s.st_size - 1, 1, f);
			printf("Status: 200 OK\r\n");
			printf("\r\n");
			printf("{\"revision\": %s, \"code\": %d}", revision, SWINEHUB_CODE_OK);
			fclose(f);
		}else{
			printf("Status: 404 Not Found\r\n");
			printf("\r\n");
			printf("{\"error\": \"Repository not found\", \"code\": %d}", SWINEHUB_CODE_REPO_NOT_FOUND);
		}
		free(str);
		free(rev);
	}else if(strcmp(path, "/v1/repo/latest-log") == 0){
		int i;
		char* backup = strdup(swinehub_body);
		for(i = 0; swinehub_body[i] != 0; i++){
			if(swinehub_body[i] == ':'){
				swinehub_body[i] = 0;
				break;
			}
		}
		char* username = swinehub_body;
		char* reponame = swinehub_body + i + 1;
		if(strlen(username) == strlen(backup)){
			free(backup);
			printf("Status: 400 Bad Request\r\n");
			printf("\r\n");
			printf("{\"error\": \"Required parameters missing, `username:reponame`\", \"code\": %d}", SWINEHUB_CODE_PARAM_MISSING);
			return;
		}
		free(backup);
		char* chdirpath = sstrcat5(config->cachepath, "/", username, "-", reponame);
		char* repopath = sstrcat5(config->path, "/", username, "-", reponame);
		struct stat st;
		if(stat(repopath, &st) != 0){
			free(repopath);
			printf("Status: 404 Not Found\r\n");
			printf("\r\n");
			printf("{\"error\": \"Repository not found\", \"code\": %d}", SWINEHUB_CODE_REPO_NOT_FOUND);
			return;
		}
		free(repopath);
		printf("Status: 200 OK\r\n");
		printf("\r\n");
		fflush(stdout);
		if(stat(chdirpath, &st) != 0){
			int pid = fork();
			if(pid == 0){
				int nullfd = open("/dev/null", O_RDWR);
				dup2(nullfd, fileno(stdout));
				dup2(nullfd, fileno(stderr));
				chdir(config->cachepath);
				char* _filepath = sstrcat5("file:///", config->path, "/", username, "-");
				char* filepath = sstrcat(_filepath, reponame);
				free(_filepath);
				execl(config->svn, config->svn, "co", filepath, NULL);
				_exit(-1);
			}else{
				waitpid(pid, 0, 0);
			}
		}
		int pid = fork();
		if(pid == 0){
			int nullfd = open("/dev/null", O_RDWR);
			dup2(nullfd, fileno(stdout));
			dup2(nullfd, fileno(stderr));
			chdir(chdirpath);
			execl(config->svn, config->svn, "up", NULL);
			_exit(-1);
		}else{
			waitpid(pid, 0, 0);
		}
		pid = fork();
		if(pid == 0){
			int nullfd = open("/dev/null", O_RDWR);
			dup2(nullfd, fileno(stderr));
			chdir(chdirpath);
			execl(config->svn, config->svn, "log", "-r", "HEAD", NULL);
			_exit(-1);
		}else{
			waitpid(pid, 0, 0);
		}
		free(chdirpath);
		fflush(stdout);
	}else if(strcmp(path, "/v1/repo/set-readme") == 0){
		datum key;
		key.dptr = getenv("HTTP_X_TOKEN");
		key.dsize = strlen(getenv("HTTP_X_TOKEN")) + 1;
		datum result = dbm_fetch(swinehub_token_db, key);
		if(result.dptr != NULL){
			char* repopath = sstrcat6(config->path, "/", (char*)result.dptr, "-", getenv("HTTP_X_REPONAME"), "/README.txt");
			FILE* f = fopen(repopath, "w");
			if(f != NULL){
				_FWRITE(f, swinehub_body);
				fclose(f);
				printf("Status: 200 OK\r\n");
				printf("\r\n");
				printf("{\"message\": \"README updated successfully\", \"code\": %d}", SWINEHUB_CODE_OK);
			}else{
				printf("Status: 404 Not Found\r\n");
				printf("\r\n");
				printf("{\"error\": \"Repository not found\", \"code\": %d}", SWINEHUB_CODE_REPO_NOT_FOUND);
			}
			free(repopath);
		}else{
			printf("Status: 403 Forbidden\r\n");
			printf("\r\n");
			printf("{\"error\": \"Invalid token\", \"code\": %d}", SWINEHUB_CODE_INVALID_TOKEN);
		}
	}else if(strcmp(path, "/v1/user/delete") == 0){
		datum key;
		key.dptr = swinehub_body;
		key.dsize = strlen(swinehub_body) + 1;
		datum result = dbm_fetch(swinehub_token_db, key);
		if(result.dptr != NULL){
			char* username = strdup(result.dptr);
			key.dptr = result.dptr;
			key.dsize = strlen(result.dptr) + 1;
			datum d_user = dbm_fetch(swinehub_user_db, key);
			struct swinehub_user* s_user = (struct swinehub_user*)malloc(sizeof(*s_user));
			memcpy(s_user, d_user.dptr, sizeof(*s_user));
			if(0 != dbm_delete(swinehub_user_db, key)){
				printf("Status: 500 Internal Server Error\r\n");
				printf("\r\n");
				printf("{\"error\": \"Database update failed: %s\", \"code\": %d}", strerror(errno), SWINEHUB_CODE_DB);
			}else{
				swinehub_db_init();
				if(s_user->isteam == 1){
					char* data = malloc(1);
					data[0] = 0;
					FILE* f = fopen(SWINEHUB_AUTHZ, "r");
					flockfile(f);
					char* linebuf = malloc(1);
					linebuf[0] = 0;
					char* buf = malloc(2);
					buf[1] = 0;
					char cutit = 0;
					char* cutfrom = sstrcat("#%START-TEAM ", username);
					while(1){
						fread(buf, 1, 1, f);
						if(feof(f) || buf[0] == '\n'){
							if(strcmp(linebuf, cutfrom) == 0){
								cutit = 1;
							}
							if(cutit == 0){
								if(strlen(linebuf) != 0){
									char* tmp = data;
									data = sstrcat3(data, linebuf, "\n");
									free(tmp);
								}
							}
							if(strcmp(linebuf, "#%END-TEAM") == 0 && cutit == 1){
								cutit = 0;
							}
							free(linebuf);
							linebuf = malloc(1);
							linebuf[0] = 0;
							if(feof(f)) break;
						}else{
							char* tmp = linebuf;
							linebuf = sstrcat(linebuf, buf);
							free(tmp);
						}
					}
					free(cutfrom);
					free(buf);
					free(linebuf);
					fclose(f);
					f = fopen(SWINEHUB_AUTHZ, "w");
					_FWRITE(f, data);
					funlockfile(f);
					fclose(f);
					free(data);
				}
				key.dptr = swinehub_body;
				key.dsize = strlen(swinehub_body) + 1;
				if(0 != dbm_delete(swinehub_token_db, key)){
					printf("Status: 500 Internal Server Error\r\n");
					printf("\r\n");
					printf("{\"error\": \"Database update failed: %s\", \"code\": %d}", strerror(errno), SWINEHUB_CODE_DB);
				}else{
					swinehub_db_init();
					DIR* dir = opendir(config->path);
					if(dir != NULL){
						struct dirent* d;
						char* cmp = sstrcat(username, "-");
						while((d = readdir(dir)) != NULL){
							if(strcmp(d->d_name, ".") == 0 || strcmp(d->d_name, "..") == 0) continue;
							if(strlen(d->d_name) < strlen(cmp)) continue;
							char* tmp = strdup(d->d_name);
							tmp[strlen(cmp)] = 0;
							if(strcmp(tmp, cmp) == 0){
								FILE* f = fopen(SWINEHUB_AUTHZ, "r");
								flockfile(f);
								char* cutfrom = sstrcat("#%START ", d->d_name);
								char* buf = malloc(2);
								buf[1] = 0;
								char* linebuf = malloc(1);
								linebuf[0] = 0;
								char* data = malloc(1);
								data[0] = 0;
								char cutit = 0;
								while(1){
									fread(buf, 1, 1, f);
									if(feof(f) || buf[0] == '\n'){
										if(strcmp(linebuf, cutfrom) == 0){
											cutit = 1;
										}
										if(strlen(linebuf) != 0 && cutit != 1){
											char* tmp = data;
											data = sstrcat3(data, linebuf, "\n");
											free(tmp);
										}
										if(strcmp(linebuf, "#%END") == 0 && cutit == 1){
											cutit = 0;
										}
										free(linebuf);
										linebuf = malloc(1);
										linebuf[0] = 0;
										if(feof(f)) break;
									}else{
										char* tmp = linebuf;
										linebuf = sstrcat(linebuf, buf);
										free(tmp);
									}
								}
								free(linebuf);
								free(buf);
								free(cutfrom);
								funlockfile(f);
								fclose(f);
								f = fopen(SWINEHUB_AUTHZ, "w");
								flockfile(f);
								_FWRITE(f, data);
								funlockfile(f);
								fclose(f);
								free(data);
								int pid = fork();
								if(pid == 0){
									chdir(config->path);
									execl("/bin/rm", "/bin/rm", "-rf", d->d_name, NULL);
									_exit(-1);
								}else{
									waitpid(pid, 0, 0);
								}
							}
							free(tmp);
						}
						free(cmp);
						closedir(dir);
					}
					dir = opendir(config->cachepath);
					if(dir != NULL){
						struct dirent* d;
						char* cmp = sstrcat(username, "-");
						while((d = readdir(dir)) != NULL){
							if(strcmp(d->d_name, ".") == 0 || strcmp(d->d_name, "..") == 0) continue;
							if(strlen(d->d_name) < strlen(cmp)) continue;
							char* tmp = strdup(d->d_name);
							tmp[strlen(cmp)] = 0;
							if(strcmp(tmp, cmp) == 0){
								int pid = fork();
								if(pid == 0){
									chdir(config->cachepath);
									execl("/bin/rm", "/bin/rm", "-rf", d->d_name, NULL);
									_exit(-1);
								}else{
									waitpid(pid, 0, 0);
								}
							}
							free(tmp);
						}
						free(cmp);
						closedir(dir);
					}
					printf("Status: 200 OK\r\n");
					printf("\r\n");
					printf("{\"message\": \"User deleted successfully\", \"code\": %d}", SWINEHUB_CODE_OK);
				}
			}
		}else{
			printf("Status: 403 Forbidden\r\n");
			printf("\r\n");
			printf("{\"error\": \"Invalid token\", \"code\": %d}", SWINEHUB_CODE_INVALID_TOKEN);
		}
	}else if(strcmp(path, "/v1/repo/log") == 0){
		int i;
		char* backup = strdup(swinehub_body);
		for(i = 0; swinehub_body[i] != 0; i++){
			if(swinehub_body[i] == ':'){
				swinehub_body[i] = 0;
				break;
			}
		}
		char* username = swinehub_body;
		char* reponame = swinehub_body + i + 1;
		char* logpath;
		if(strlen(username) == strlen(backup)){
			free(backup);
			printf("Status: 400 Bad Request\r\n");
			printf("\r\n");
			printf("{\"error\": \"Required parameters missing, `username:reponame:path`\", \"code\": %d}", SWINEHUB_CODE_PARAM_MISSING);
			return;
		}
		for(i = 0; reponame[i] != 0; i++){
			if(reponame[i] == ':'){
				reponame[i] = 0;
				break;
			}
		}
		logpath = reponame + i + 1;
		if(strlen(reponame) + 1 + strlen(username) == strlen(backup)){
			free(backup);
			printf("Status: 400 Bad Request\r\n");
			printf("\r\n");
			printf("{\"error\": \"Required parameters missing, `username:reponame:path`\", \"code\": %d}", SWINEHUB_CODE_PARAM_MISSING);
			return;
		}
		free(backup);
		char* chdirpath = sstrcat5(config->cachepath, "/", username, "-", reponame);
		char* repopath = sstrcat5(config->path, "/", username, "-", reponame);
		struct stat st;
		if(stat(repopath, &st) != 0){
			free(repopath);
			printf("Status: 404 Not Found\r\n");
			printf("\r\n");
			printf("{\"error\": \"Repository not found\", \"code\": %d}", SWINEHUB_CODE_REPO_NOT_FOUND);
			return;
		}
		free(repopath);
		printf("Status: 200 OK\r\n");
		printf("\r\n");
		fflush(stdout);
		if(stat(chdirpath, &st) != 0){
			int pid = fork();
			if(pid == 0){
				int nullfd = open("/dev/null", O_RDWR);
				dup2(nullfd, fileno(stdout));
				dup2(nullfd, fileno(stderr));
				chdir(config->cachepath);
				char* _filepath = sstrcat5("file:///", config->path, "/", username, "-");
				char* filepath = sstrcat(_filepath, reponame);
				free(_filepath);
				execl(config->svn, config->svn, "co", filepath, NULL);
				_exit(-1);
			}else{
				waitpid(pid, 0, 0);
			}
		}
		int pid = fork();
		if(pid == 0){
			int nullfd = open("/dev/null", O_RDWR);
			dup2(nullfd, fileno(stdout));
			dup2(nullfd, fileno(stderr));
			chdir(chdirpath);
			execl(config->svn, config->svn, "up", NULL);
			_exit(-1);
		}else{
			waitpid(pid, 0, 0);
		}
		char* filepath = sstrcat("./", logpath);
		pid = fork();
		if(pid == 0){
			int nullfd = open("/dev/null", O_RDWR);
			dup2(nullfd, fileno(stderr));
			chdir(chdirpath);
			execl(config->svn, config->svn, "log", "-r", "COMMITTED", filepath, NULL);
			_exit(-1);
		}else{
			waitpid(pid, 0, 0);
		}
		free(filepath);
		free(chdirpath);
	}else if(strcmp(path, "/v1/repo/create") == 0){
		int i;
		char* backup = strdup(swinehub_body);
		for(i = 0; swinehub_body[i] != 0; i++){
			if(swinehub_body[i] == ':'){
				swinehub_body[i] = 0;
				break;
			}
		}
		char* token = swinehub_body;
		char* reponame = swinehub_body + i + 1;
		if(strlen(token) == strlen(backup)){
			free(backup);
			printf("Status: 400 Bad Request\r\n");
			printf("\r\n");
			printf("{\"error\": \"Required parameters missing, `token:reponame`\", \"code\": %d}", SWINEHUB_CODE_PARAM_MISSING);
			return;
		}
		free(backup);
		datum key;
		key.dptr = token;
		key.dsize = strlen(token) + 1;
		datum result = dbm_fetch(swinehub_token_db, key);
		if(result.dptr != NULL){
			char* actualpath = sstrcat5(config->path, "/", result.dptr, "-", reponame);
			struct stat st;
			if(stat(actualpath, &st) == 0){
				printf("Status: 403 Forbidden\r\n");
				printf("\r\n");
				printf("{\"error\": \"Repository already exists\", \"code\": %d}", SWINEHUB_CODE_REPO_ALREADY_EXISTS);
				free(actualpath);
				return;
			}
			free(actualpath);
			FILE* authz = fopen(SWINEHUB_AUTHZ, "a");
			flockfile(authz);
			char* name = sstrcat4(result.dptr, "-", reponame, ":/");
			key.dptr = strdup(result.dptr);
			key.dsize = strlen(result.dptr) + 1;
			datum user = dbm_fetch(swinehub_user_db, key);
			struct swinehub_user* s_user = (struct swinehub_user*)user.dptr;
			char* data;
		       	if(s_user->isteam == 1){
				data = sstrcat5("\n[", name, "]\n* = r\n@", key.dptr, " = rw\n");
			}else{
				data = sstrcat5("\n[", name, "]\n* = r\n", key.dptr, " = rw\n");
			}
			fflush(stderr);
			char* tmp = data;
			data = sstrcat6("\n#%START ", key.dptr, "-", reponame, data, "#%END\n");
			free(tmp);
			free(name);
			_FWRITE(authz, data);
			funlockfile(authz);
			fclose(authz);
			free(data);

			data = sstrcat3(key.dptr, "-", reponame);
			int pid = fork();
			if(pid == 0){
				chdir(config->path);
				execl(config->svnadmin, config->svnadmin, "create", data, NULL);
				_exit(-1);
			}else{
				waitpid(pid, 0, 0);
			}
			printf("Status: 200 OK\r\n");
			printf("\r\n");
			printf("{\"message\": \"Repository successfully created\", \"code\": %d}", SWINEHUB_CODE_OK);
		}else{
			printf("Status: 403 Forbidden\r\n");
			printf("\r\n");
			printf("{\"error\": \"Invalid token\", \"code\": %d}", SWINEHUB_CODE_INVALID_TOKEN);
		}
	}else if(strcmp(path, "/v1/repo/delete") == 0){
		int i;
		char* backup = strdup(swinehub_body);
		for(i = 0; swinehub_body[i] != 0; i++){
			if(swinehub_body[i] == ':'){
				swinehub_body[i] = 0;
				break;
			}
		}
		char* token = swinehub_body;
		char* reponame = swinehub_body + i + 1;
		if(strlen(token) == strlen(backup)){
			free(backup);
			printf("Status: 400 Bad Request\r\n");
			printf("\r\n");
			printf("{\"error\": \"Required parameters missing, `token:reponame`\", \"code\": %d}", SWINEHUB_CODE_PARAM_MISSING);
			return;
		}
		free(backup);
		datum key;
		key.dptr = token;
		key.dsize = strlen(token) + 1;
		datum result = dbm_fetch(swinehub_token_db, key);
		if(result.dptr != NULL){
			char* actualpath = sstrcat5(config->path, "/", result.dptr, "-", reponame);
			struct stat st;
			if(stat(actualpath, &st) != 0){
				printf("Status: 404 Not Found\r\n");
				printf("\r\n");
				printf("{\"error\": \"Repository not found\", \"code\": %d}", SWINEHUB_CODE_REPO_NOT_FOUND);
				free(actualpath);
				return;
			}
			free(actualpath);
			FILE* authz = fopen(SWINEHUB_AUTHZ, "r");
			flockfile(authz);
			char* authzbuf = malloc(1);
			authzbuf[0] = 0;
			char* buf = malloc(1);
			buf[0] = 0;
			char ch;
			int length = 0;
			char cut_str = 0;
			char* start_tag = sstrcat4("#%START ", result.dptr, "-", reponame);
			while(1){
				fread(&ch, 1, 1, authz);
				if(feof(authz) || ch == '\n'){
					if(strcmp(buf, start_tag) == 0){
						cut_str = 1;
					}
					if(cut_str != 1){
						if(strlen(buf) != 0){
							char* tmp = authzbuf;
							authzbuf = sstrcat3(tmp, buf, "\n");
							free(tmp);
						}
					}
					if(strcmp(buf, "#%END") == 0){
						cut_str = 0;
					}
					free(buf);
					buf = malloc(1);
					buf[0] = 0;
					length = 0;
					if(feof(authz)) break;
				}else{
					buf = realloc(buf, length + 2);
					buf[length] = ch;
					buf[length + 1] = 0;
					length++;
				}
			}
			free(buf);
			fclose(authz);
			authz = fopen(SWINEHUB_AUTHZ, "w");
			_FWRITE(authz, authzbuf);
			funlockfile(authz);
			fclose(authz);
			free(authzbuf);
			char* repofull = sstrcat3(result.dptr, "-", reponame);
			int pid = fork();
			if(pid == 0){
				chdir(config->path);
				execl("/bin/rm", "/bin/rm", "-rf", repofull, NULL);
				_exit(-1);
			}else{
				waitpid(pid, 0, 0);
			}
			free(repofull);
			printf("Status: 200 OK\r\n");
			printf("\r\n");
			printf("{\"message\": \"Repository successfully deleted\", \"code\": %d}", SWINEHUB_CODE_OK);
		}else{
			printf("Status: 403 Forbidden\r\n");
			printf("\r\n");
			printf("{\"error\": \"Invalid token\", \"code\": %d}", SWINEHUB_CODE_INVALID_TOKEN);
		}
	}else{
		printf("Status: 400 Bad Request\r\n");
		printf("\r\n");
		printf("{\"error\": \"Unknown request\", \"code\": %d}", SWINEHUB_CODE_UNKNOWN_REQUEST);
	}
}
